#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id: guieditor.tcl,v 1.8 2000/10/15 17:19:20 artur Exp $
#
# Information:
# Modul (benutzt von Form Editor)
# Formengine ist zustndigt fr ein
# Drag & Drop Editieren von Formularen
Class GuiEditor -parameter {
    form_node
}
GuiEditor instproc init args {
    [self] instvar win selectedObj drag_object newwidget
    regexp {.*::(.+)} [self] _ win
    set win .$win
    set selectedObj {}
    set drag_object {}
    set newwidget {}
    [self] build_window
    [self] initwidgetset
}
# Baut das Haupteditierfenster
# Attribut win mu initialisiert sein
GuiEditor instproc build_window {} {
    [self] instvar win depositorywin workbench dataset_win widgetset \
	red_style green_style
    global lan
    toplevel $win
    wm withdraw $win
    menu $win.mb -tearoff 0
    menu $win.mb.edit -tearoff 0
    menu $win.mb.tools -tearoff 0

    $win.mb add cascade -label $lan(menu.edit.edit) -menu $win.mb.edit \
	-underline 0
    $win.mb add cascade -label $lan(menu.tools.tools) -menu $win.mb.tools \
	-underline 0
    $win.mb.edit add command -label $lan(menu.edit.formproperties) -command \
	"[self] editFormProperties"
    $win.mb.edit add command -label $lan(menu.edit.endedit) -command \
	"[self] end_edit"
    tixScrolledWindow $win.swdepository
    set depositorywin [$win.swdepository subwidget window]
    tixScrolledWindow $win.swworkbench
    set workbench [$win.swworkbench subwidget window]
    label $win.statusbar -text statusbar -relief raised -border 5
    frame $win.toolbar -relief raised -border 5
    tixNoteBook $win.notebook
    $win.notebook add widgetset -label $lan(guiedit.widgetset)
    $win.notebook add dataset_win -label $lan(guiedit.datapool)
    set f [$win.notebook subwidget widgetset]
    set widgetset [tixScrolledTList $f.list -options {
	tlist.orient horizontal
	tlist.selectMode single
	tlist.command {}
	tlist.width 20
	tlist.height 20
    }
		  ]
    set f [$win.notebook subwidget dataset_win]
    set dataset_win [tixScrolledTList $f.list -options {
	tlist.orient horizontal
	tlist.selectMode single
	tlist.command {}
	tlist.width 20
	tlist.height 20
    }
		 ]
    pack $dataset_win -expand yes -fill both
    pack $widgetset -expand yes -fill both
    # DUMMY button
    # button $win.toolbar.b1 -text button
    # Image Command
    set toolcommand {
	{editcut {doSelectedObject moveDepository}}
	{eraser {doSelectedObject deleteObject}}
	{up {doSelectedObject selectparent}}
	{next {doSelectedObject selectnext}}
	{prev {doSelectedObject selectprev}}
    }
    foreach tool $toolcommand {
	set image [lindex $tool 0]
        set name [lindex $tool 1]
	button $win.toolbar.$name -image [tix getimage $image] \
	    -command "[self] $name" \
            -takefocus 0 -bg #c0c0c0 -highlightbackground #c0c0c0
        pack $win.toolbar.$name -side left -anchor w
	$win.mb.tools add command -label $lan(menu.tools.${image}) \
		    -command "[self] $name"
    }
    grid $win.toolbar -row 0 -column 0 -columnspan 2 -sticky we
    grid $win.notebook -row 1 -column 0 -sticky ns
    grid $win.swworkbench -row 1 -column 1 -sticky wnse
    grid $win.swdepository -row 2 -column 0 -columnspan 2 -sticky wesn
    grid $win.statusbar -row 3 -column 0 -columnspan 2 -sticky we
    grid rowconfigure $win 1 -weight 1
    grid columnconfigure $win 1 -weight 1 -minsize 500
    grid rowconfigure $win 2 -minsize 100
    $win configure -menu $win.mb 
    # Standart Window Close Abfangen
    wm protocol $win WM_DELETE_WINDOW "[self] end_edit"
    set red_style  [tixDisplayStyle text -fg red]
    set green_style [tixDisplayStyle text -fg green4]
 }
# Eine PostInitialisierung, gedacht um nach der Init der abgeleiteten
# Klasse abgerufen zu werden
GuiEditor instproc postInit {} {
    [self] instvar win
    bind $win <ButtonRelease-1> "[self] try_enddrag %W %X %Y"
    wm deiconify $win
}
# Initiert die Auswahlmenge von unterttzten Widget
# im tixTree das im widgetset gesetzt ist
GuiEditor instproc initwidgetset {} {
    [self] instvar widgetset widgetlist
    set hlist [$widgetset subwidget tlist]
    # Problem die Reihenfolge ist zufllig
    foreach ewidget [EditWidget info subclass] {
	$hlist insert end -itemtype imagetext \
	    -text [$ewidget set name]  -image [tix getimage [$ewidget set image]]
	    #-data $ewidget
	lappend widgetlist $ewidget
    }
    #$hlist configure -command "[self] new_widget"
    bind $hlist <B1-Leave> "[self] new_widget"
}
GuiEditor instproc end_edit {} {
    [self] instvar win
    #puts "GUI EDIT end_edit"
    destroy $win
    [self] destroy
}
GuiEditor instproc new_widget {} {
    [self] instvar widgetlist win newwidget widgetset
    set hlist [$widgetset subwidget tlist]
    set index [$hlist info selection]
    if {$index==""} {return}
    $win configure -cursor target
    set newwidget [lindex $widgetlist $index]
    [self] setStatusInfo "Definiere das Ziel fr Widget $newwidget"
}
GuiEditor instproc selectObject {obj} {
    [self] instvar selectedObj win
    foreach sobj $selectedObj {
	$sobj unselect
    }
    $obj select
    set selectedObj $obj
    # Wenn ein Options Window geffnet ist. Dann benutzt es fr dieses Object
    if {[winfo exist .options]} {
	$obj editAttr $win
    }
}
GuiEditor instproc selectAddObject {obj} {
    [self] instvar selectedObj
    $obj select
    lappend selectedObj $obj
}
GuiEditor instproc moveDepository {obj} {
    [self] instvar depositorywin
    # Wenn das Objekt bis jetzt 'unmapped' ist
    # z.B wenn es gereade kreirt ist
    if {[info exist ${obj}::win] && [$obj set win]!=""} {
	$obj release
	set winname [winfo name [$obj set win]]
	$obj unmap
    } else {
	set winname [Object autoname widget]
    }
    $obj map $depositorywin.$winname
    [self]::depository packObject $obj {-side left}
    $obj foreach_do "[self] bindObject"
}
# Hier hat die Operation sinn wenn ein Target per Maus fr alle
# Depository Objekte Spezifiezirt wird
GuiEditor instproc copyDepository {obj} {
    #$obj release
}
GuiEditor instproc deleteObject {obj} {
    [self] instvar selectedObj datawidget_arr
    if {[[$obj info class] set typ]=="container" && [llength [$obj set widgetlist]]>0} {
	global lan
^        tk_messageBox -message $lan(message.notemptycontainer) -icon warning -type ok
        return
    } 
    set index [lsearch $selectedObj $obj]
    if {$index>=0} {
	set selectedObj [lreplace $selectedObj $index $index] 
	#puts "deleted $obj $selectedObj"
    }
    if {[info exist ${obj}::data] && [lcontain [keylkeys ${obj}::data] data]} {
	if {[info exist datawidget_arr([keylget ${obj}::data data])]} {
	    [self] dataset_minus $obj [keylget ${obj}::data data]
	}
    }
    $obj release
    $obj unmap
    $obj destroy
}
GuiEditor instproc selectparent {obj} {
    [self] instvar datawidget_arr
    if {[set newobj [$obj getParent]]!=""} {
	[self] selectObject $newobj
	[self] enterObject $newobj
    }
}
GuiEditor instproc selectnext {obj} {
    [self] instvar selectedObj
    if {[set newobj [$obj getNext]]!=""} {
	[self] selectObject $newobj
	[self] enterObject $newobj
    }
}
GuiEditor instproc selectprev {obj} {
    [self] instvar selectedObj
    if {[set newobj [$obj getPrev]]!=""} {
	[self] selectObject $newobj
	[self] enterObject $newobj
    }
}
GuiEditor instproc morfObject {obj morfclass {transAttr {}}} {
    # Es lebe hoch XOTCL
    # Dynamische Klassennderung
    #puts "moorf $obj $morfclass $transAttr"
    $obj class $morfclass
    # Initialisiere Defaultwerte die jetzt leer sind
    foreach oclass [concat $morfclass [$morfclass info superclass]] {
	foreach par [$oclass info parameter] {
	    if {[$obj info vars [lindex $par 0]]=="" && [llength $par]>1} {
		$obj set [lindex $par 0] [lindex $par 1]
	    }
	}
    }
    foreach attr $transAttr {
	$obj set [lindex $attr 1] [$obj set [lindex $attr 0]]
    }
    set packoptd [[set ${obj}::geo_manager] getOptions $obj]
    foreach opt $packoptd {
	keylset packopt [keylget opt name] [keylget opt default]
    }
    $obj unmap
    $obj map [$obj set win]
    [set ${obj}::geo_manager] setOptions $obj $packopt
    [self] bindObject $obj
}
#
# Toolbars und Mens entrypoints
#
GuiEditor instproc doSelectedObject {command} {
    [self] instvar selectedObj 
    foreach obj $selectedObj {
	$command $obj
    }
}
GuiEditor instproc popmenuObject {obj x y} {
    [self] instvar win
    global lan
    set m $win.popmenu
    destroy $m
    
    menu $m -tearoff 0
    if {"[$obj set geo_manager]"!="[self]::depository"} {
	$m add command -label $lan(popmenu.cut) \
	    -command "[self] moveDepository $obj"
    }
    #$m add command -label $lan(popmenu.copy) \
 	-command "[self] copyDepository $obj"
    $m add command -label $lan(popmenu.delete) \
	-command "[self] deleteObject $obj"
    set oclass [$obj info class]
    if {[$oclass info vars morf]!=""} {
	menu $m.morf -tearoff 0
	foreach morfinfo [$oclass set morf] {
	    set morfclass [lindex $morfinfo 0]
	    $m.morf add command -label [$morfclass set name] \
		-command "[self] morfObject $obj $morfclass [list [lrange $morfinfo 1 end]]"
	}
	$m add cascade -menu $m.morf -label $lan(popmenu.morf)
    }
    $m add separator
    $m add command -label $lan(popmenu.options) \
	-command "$obj editAttr $win"
    tk_popup $m $x $y
}

# Benutzt um die Binding fr ein Object zu schaffen
# bei neuen Object
GuiEditor instproc bindObject {object} {
    [self] instvar win
    set cwin [$object set win]
    bind $cwin <Enter> "[self] enterObject $object"
    #Bind fr Eigenschafte Double-Button-1 wird bereits in dem Objekt
    #erledigt, da wird auch entsprechen die bindtags umgelaeitet
    bind $cwin <ButtonPress-1> "[self] selectObject $object"
    bind $cwin <ButtonPress-3> "[self] popmenuObject $object %X %Y"
    bind $cwin <Shift-ButtonPress-1> "[self] selectAddObject $object"
    # Drag&Drop wird realisiert wenn bei gepressten Button1 
    # das Widget verlassen wird
    # andere Methode specTcl (button_bind.tk) bei Motion
    bind $cwin <B1-Leave>  "[self] start_drag $object"
}
GuiEditor instproc start_drag {object} {
    [self] instvar drag_object win
    set drag_object $object
    $win configure -cursor hand2
}
# Diese Prozedur wird am Ende der Drag&Drop Operation ausgefhrt
# loslassen der Maustaste. Es wird das Ziel ermittelt und die entsprechende
# command fr Bewegung Operation wird mittels 'searchDropCommand'
# abgefragt
GuiEditor instproc try_enddrag {dwin posx posy} {
    [self] instvar win drag_object depositorywin workbench rootframe newwidget
    #puts "try_enddrag $dwin $posx $posy"
    if {$drag_object==""} {
	# Neues Widget Einfgen
	if {$newwidget!=""} {
	    [self] try_newwidget $dwin $posx $posy
	}
	return
    }
    #puts "$dwin $posx $posy well"
    $win configure -cursor {}
    if {[pointInWindow $dwin $posx $posy]} then {
	# Nichts Das Objekt wurde nicht bewegt (Maus-Button
	# wieder ber das Objekt losgelasssen
	#puts "No Drag&Drop"
    } elseif {[pointInWindow $depositorywin $posx $posy] &&
	![string match *depository* $dwin]} then {
	# Das Objekt wurde in die Ablage verschoben
	# und es war nicht in der Ablage
	[self] moveDepository $drag_object
    } elseif {[pointInWindow $workbench $posx $posy]} {
	set command [$rootframe searchDropCommand $posx $posy]
	#puts "Drop command $command"
	if {$command!=""} {
	    $drag_object release
	    if {[string match *depository* [$drag_object set win]]} {
		# Wenn sich der Object in der Anlage befindet mu
		# das Window name neu gemacht werden
		set winname [winfo name [$drag_object set win]]
		$drag_object unmap
		$drag_object map $workbench.$winname
		$drag_object foreach_do "[self] bindObject"
	    }
	    eval [format $command $drag_object]
	}
    }
    set drag_object {}
}
# Aufgerufen von ButtonPress Callback 
# Sucht das Zeil fr Einfgen von neuem Widget
GuiEditor instproc try_newwidget {dwin posx posy} {
    [self] instvar newwidget win workbench depositorywin rootframe
    #puts "end_newwidget"
    if {$newwidget==""} { return }
    if {[pointInWindow $depositorywin $posx $posy]} then {
	# Das Objekt soll in die Ablage entstehen
	set widget [EditWidget makeWidget $newwidget [self]::[Object autoname widget]]
	[self] moveDepository $widget
    } elseif {[pointInWindow $workbench $posx $posy]} {
	set command [$rootframe searchDropCommand $posx $posy]
	#puts "Drop command $command"
	if {$command!=""} {
	    set name [Object autoname widget]
	    set widget [EditWidget makeWidget $newwidget [self]::$name]
	    $widget map $workbench.$name
	    [self] bindObject $widget
	    eval [format $command $widget]
	    if {[info exist ${widget}::data]} {
		[self] dataset_add $widget [keylget ${widget}::data data]
	    }
	}
    }
    $win configure -cursor {}
    set newwidget {}
}
#
# Diese Klasse implementiert XML und DB Formular spezifische Funktionnen
# auch die Erzeugung von Widget und Packer
#
Class DBGuiEditor -superclass GuiEditor -parameter {
    {schema_node {}}
}
DBGuiEditor instproc init args {
    [self] instvar win datawidget_arr form_node schema_node workbench depositorywin rootframe
    next
    # Windows Titel setzen
    if {[$form_node hasAttribute name]} {
	wm title $win [$form_node getAttribute name]
    }
    #set datawidget_arr(test) ew
    update idletask
    grid propagate $win 0
    set rootframe [mffactory makeeditable_fromdom $form_node [self]]
    array set datawidget_arr [array get mffactory::data]
    #puts "DATAWIDGET [array names datawidget_arr]"
    $rootframe map $workbench
    # Initiere Depository (Ablage) als EPackFrame
    EPackFrame [self]::depository    
    [self]::depository set win $depositorywin
    [self]::depository set create_id -1
    # Setzt die Bindings fr alle Widget
    # auer workbench und depositorywin
    foreach child [[self] info children] {
	if {[info exist ${child}::win]} {
	    set cwin [set ${child}::win]
	    # Ignoriere Workbench Window
	    if {$cwin==$workbench || $cwin==$depositorywin} { continue }
	    [self] bindObject $child
	}
    }
    [self] initDataSet
    [self] postInit
}
#
# Initiealisiere Daten Auswahl Liste
DBGuiEditor instproc initDataSet {} {
    [self] instvar dataset_win datalist alldataKL datawidget_arr green_style red_style
    set tlist [$dataset_win subwidget tlist]
    [self] searchAllData
    set datalist {}
    # Problem die Reihenfolge ist zufllig
    foreach data [array names datawidget_arr] {
	#puts "adding $data"
	if {$alldataKL==""} {
	    $tlist insert end -itemtype text \
		-text $data
	    lappend datalist $data
	} else {
	    if {[lcontain [keylkeys alldataKL] $data]} {
		$tlist insert end -itemtype text \
		    -text "$data ([[keylget alldataKL $data] getAttribute type])"
		lappend datalist $data
	    } else {
		$tlist insert end -itemtype text \
		    -text "$data" -style $red_style
		lappend datalist $data
	    }
	}
    }
    if {$alldataKL!=""} {
	foreach data [keylkeys alldataKL] {
	    if {![lcontain $datalist $data]} {
		$tlist insert end -itemtype text \
		    -text "$data ([[keylget alldataKL $data] getAttribute type])" -style $green_style
		lappend datalist $data
	    }
	}
    }
    $tlist configure -command "[self] selectByData"
}
DBGuiEditor instproc dataset_add {obj data} {
    [self] instvar datalist win dataset_win datawidget_arr alldataKL \
	red_style
    set tlist [$dataset_win subwidget tlist]
    if {[info exist datawidget_arr($data)]} {
	lappend datawidget_arr($data) $obj
    } else {
	if {$alldataKL==""} {
	    $tlist insert end -itemtype text \
		-text $data
	    lappend datalist $data
	} else {
	    if {[lcontain [keylkeys alldataKL] $data]} {
		# Voraussetzung es wird bereit angezeigt aber grnn
		set id [lsearch $datalist $data]
		$tlist entryconfigure $id -style {}
	    } else {
		$tlist insert end -itemtype text \
		    -text $data -style $red_style
		lappend datalist $data
	    }
	}
	set datawidget_arr($data) $obj
    }
}
DBGuiEditor instproc dataset_minus {obj data} {
    [self] instvar datalist win dataset_win datawidget_arr alldataKL green_style
    set tlist [$dataset_win subwidget tlist]
    if {[info exist datawidget_arr($data)]} {
	set actobjs $datawidget_arr($data)
	if {[llength $actobjs]==1} {
	    unset datawidget_arr($data)
	    set id [lsearch $datalist $data]
	    if {$id<0} {
		error {Zustand inkosistent}
	    }
	    if {$alldataKL=="" || ![lcontain [keylkeys alldataKL] $data]} {
		set datalist [lreplace $datalist $id $id]
		$tlist delete $id
	    } else {
		$tlist entryconfigure $id -style $green_style
	    }
	} else {
	    # Objekt-Eintrag aus der Liste entfernen
	    set datawidget_arr($data) [minusMenge $actobjs $obj]
	}
    }
}
DBGuiEditor instproc selectByData {index} {
    [self] instvar datalist win dataset_win datawidget_arr selectedObj
    set tlist [$dataset_win subwidget tlist]
    set index [$tlist info selection]
    if {$index==""} { return }
    set data [lindex $datalist $index]
    if {[info exist datawidget_arr($data)]} {
	if {![lcontain $datawidget_arr($data) $selectedObj]} {
	    set nobj [lindex $datawidget_arr($data) 0]
	    [self] selectObject $nobj
	} elseif {[llength [set olist $datawidget_arr($data)]]>1} {
	    set id [lsearch -glob $olist $selectedObj]
	    incr id
	    if {$id==[llength $olist]} {
		set id 0
	    }
	    [self] selectObject [lindex $olist $id]
	}
    }
}
# bertrgt die Anderungen auf Dateherkunft der Objekte auf datawidget_attr
DBGuiEditor instproc changeObject {obj newdata} {
    [self] dataset_minus $obj [keylget ${obj}::data data]
    [self] dataset_add $obj $newdata
}
# Durschucht schema nach attributen und Kompletiert eine KeydList
# key=Name des Attribut
# value=tDom node 
DBGuiEditor instproc searchAllData {} {
    [self] instvar alldataKL schema_node form_node
    if {$schema_node!=""} {
	foreach attrnode [$schema_node selectNodes child::node()\[not(self::table)\]/descendant-or-self::attr] {
	    keylset alldataKL  [$attrnode getAttribute name] $attrnode
	}
    } else {
	set alldataKL {}
    }
}
#
# Aufgerufen bei Editieren der Widget Attributen
# 
DBGuiEditor instproc getFormNames {} {
    [self] instvar form_node
    return [$form_node selectNodes //child::form/attribute::name]
}
# Information in der StatusZeile setzten
DBGuiEditor instproc setStatusInfo {text} {
    [self] instvar win
    $win.statusbar configure -text $text
}
# Diese Prozedur wird als ein Callback aufgerufen wenn ein Pointer ber ein
# Objekt sich befindet. Es wird eine Meldung in Statusleiste ausgegeben
DBGuiEditor instproc enterObject objekt {
    [self] instvar newwidget
    # Wenn Ziel fr neues Widget gesucht wird dann keine Informationen
    if {$newwidget!=""} { return }
    set statusinfo [$objekt getinfo]
    [self] setStatusInfo $statusinfo
}
DBGuiEditor instproc end_edit {} {
    [self] instvar form_node rootframe win
    global lan
    if {[tk_messageBox -title $lan(dialog.warning) \
	     -message $lan(dialog.applyformchanges) -icon question \
	     -type okcancel]=="ok"} {
	set tdom [$rootframe asTDom [$form_node ownerDocument]]
	# Hier werden die alte Form mit der neuen ersetz
	# Eigentlich nur das erste Frame die
	# Es mssten noch die neue Form Attribute hier geschrieben werden
	foreach childnode [$form_node childNodes] {
	    if {[$childnode nodeType]=="ELEMENT_NODE" && 
		[$childnode nodeName]=="frame"} {
		$form_node replaceChild $tdom $childnode
		# Wie lscht man Nodes ????
		#$childnode delete
		break
	    }
	}
	#puts [$tdom asXML]
	#puts "Vor dem Destroy"
	#global .myDBGuiEditor0.swworkbench.f2.window.myWidget0
	#puts "ARRNAMES [array names .myDBGuiEditor0.swworkbench.f2.window.myWidget0]"
    }
    destroy $win
    #puts "Nach dem Destroy"
    [self] destroy
}
# Dialog fr allgemeine FormEigenschaften
DBGuiEditor instproc editFormProperties {} {
    [self] instvar form_node schema_node
    keylset values table [$form_node getAttribute table]
    keylset values name [$form_node getAttribute name]
    keylset values primarykey [$form_node getAttribute primarykey]
    set prop1 {{type OLabelEntry} {name name} {label name} {textwidth 20}}
    set prop2 {{type OLabelEntry} {name table} {label table}}
    set prop3 {{type OLabelEntry} {name primarykey} {label primarykey} {textwidth 20}}
    if {$schema_node!=""} {
	keylset prop2 type OComboBox
	keylset prop2 valuelist [$schema_node selectNodes //descendant::table/attribute::name]
    }
    OInputDialog [self]::propdialog [list $prop1 $prop2 $prop3] $values
    if {[set ret [[self]::propdialog execute]]!=""} {
	foreach ckey [keylkeys ret] {
	    $form_node setAttribute $ckey [keylget ret $ckey]
	    if {$ckey=="name"} {
		# ndern von Anzeige wenn Namensnderung !!!
		FormEditor applyChanges $form_node
	    }
	}
    }
    [self]::propdialog destroy
}
#
# Hilfsfunktionnen
#
proc pointInWindow {win x y} {
    set rootx [winfo rootx $win]
    set rooty [winfo rooty $win]
    #puts "x=$x y=$y $rootx $rooty [winfo width $win] [winfo height $win]"
    if {$rootx<$x && [expr $rootx+[winfo width $win]]>$x && 
	$rooty<$y && [expr $rooty+[winfo height $win]]>$y} {
	return 1
    } else {
	return 0
    }
}
