#####################################################################
# Copyright (C) 2000 Artur Trzewik
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
# Header
# CVS - $Id
#
# Information:
# Modul (benutzt von Schema Editor)
# implementiert eine einfache Schema-Basierte ReverseEngineering
# In 4 Schritten
# 1. Finden von Fremsschlssel
# 2. Finden von Vererbungen
# 3. Finden von reltionship Tablelen (die n:m oder n:m:z.. Beziehungne modelieren
# 4. Vorschlagen von Aggregationen (Tabellen in Tabellen entwetde als part-of oder id-relationship)

Class ReverseEngeneering -parameter {schema_node}
ReverseEngeneering instproc map {{window .reverse}} {
    [self] instvar win textwin messagewin
    set win $window
    global lan
    catch {destroy $win}
    toplevel $win -class Dialog
    wm title $win "Reverse Engeneering [[self] info class]"
    wm iconname $win Dialog
    wm protocol $win WM_DELETE_WINDOW "[self] destroy"
    message $win.info -width 550 \
	-text $lan(re_info.[namespace tail [[self] info class]])
	
    set textwin [[tixScrolledText $win.stext] subwidget text]
    set messagewin [message $win.message]
    tixButtonBox $win.buttons -orientation horizontal
    $win.buttons add ok -text $lan(standard.ok) -command "[self] apply"
    $win.buttons add cancel -text $lan(standard.cancel) -command "[self] destroy"
    pack $win.info -fill x -expand 1
    pack $win.stext -side top -fill both -expand 1
    pack $win.buttons -side bottom -fill x -expand 1

    focus $win
}
ReverseEngeneering instproc execute {} {
    [self] instvar textwin messagewin
    [self] map
    [self] search
}
ReverseEngeneering instproc destroy {} {
    [self] instvar win
    #puts destroy
    ::destroy $win
    next
}
ReverseEngeneering instproc search {} {
    [self] instvar textwin
    global lan
    if  {[[self] info children]==""} {
	$textwin insert end $lan(re_info.nonewstruct)
    }
    $textwin configure -state disabled
}
ReverseEngeneering instproc apply {} {
    foreach obj [[self] info children] {
	$obj apply
    }
    [self] destroy
}
Class RE_foreignkeys -superclass ReverseEngeneering
RE_foreignkeys instproc search {} {
    [self] instvar schema_node textwin
    # Finde alle Attribute, die ganua so heien wie Primrschlssel der anderen
    # Tabellen abel nicht Prim-Attribut sind oder nicht 
    # einziges Primatribut sind
    
    # Finde Alle Ein Attributige Primrschlssel
    foreach table [$schema_node selectNodes descendant::table] {
	set primattr [tdom_getprimattr $table]
	set tname [$table getAttribute name]
	#puts "PA $primattr $tname"
	if {[llength $primattr]==1} {
	    lappend obj_prims($primattr) $tname
	}
    }
    # berprfe ob es Korespondierende Fremdschlssel gibt
    foreach aname [array names obj_prims] {
	# Suche die Fremdschlssel
	set fkattr [$schema_node selectNodes descendant::table/child::node()\[not(self::table)\]/descendant-or-self::attr\[@name=\"$aname\"\]]
	foreach fkey $fkattr {
	    set fktable [lindex [$fkey selectNodes ancestor::table] 0]
	    set fktable_name [$fktable getAttribute name]
	    # Referenz auf sich Selbs oder Mgliche Vererbung auschlieen
	    if {[lcontain $obj_prims($aname) $fktable_name]} {
		continue
	    }
	    # Existiert der Fremdschlssel bereit als in assoziationcollenction
	    if {[$fktable selectNodes child::assoziationcollection/child::participant\[@reference=\"$aname\"\]]!=""} {
		continue
	    }
	    # Ist es bereit als Aggregation modelliert
	    if {[$fktable hasAttribute reference] && [$fktable getAttribute reference]==$aname} {
		continue
	    }	    
	    RE_foreignkeys::Executor [self]::[Object autoname erExecutor] \
		-primattr $aname \
		-primtables $obj_prims($aname) \
		-foreigntable $fktable_name \
		-textwin $textwin
	    
	}
    }
    #puts [array get obj_prims]
    next
}
Class RE_foreignkeys::Executor -parameter {primattr primtables foreigntable textwin}
RE_foreignkeys::Executor instproc init {args} {
    [self] instvar primattr primtables foreigntable textwin check_apply primtable
    set check_apply 1
    global lan
    set wpost [namespace tail [self]]
    checkbutton $textwin.c$wpost -text $lan(re_info.accept) -variable [self]::check_apply
    $textwin window create end -window $textwin.c$wpost
    $textwin insert end " [format $lan(re_info.pkfromtable) $primattr] "
    if {[llength $primtables]>1} {
	eval tk_optionMenu $textwin.op$wpost [self]::primtable $primtables
	$textwin window create end -window $textwin.op$wpost
    } else {
	set primtable $primtables
	$textwin insert end " \"$primtable\""
    }
    $textwin insert end "\n      $lan(re_info.foundin) \"$foreigntable\"\n"
}
RE_foreignkeys::Executor instproc apply {} {
    [self] instvar primattr foreigntable check_apply primtable
    if {$check_apply} {
	# !!! Voraussetzung parent hat Attribut schema_node
	set domNode [[[self] info parent] set schema_node]
	# generiere zuert assoziationtarget
	set t_node  [$domNode selectNodes \
			 descendant::table\[@name=\"$primtable\"\]]
	set at_node [[$domNode ownerDocument] createElement assoziationtarget]
	$at_node setAttribute name ${primtable}$foreigntable
	$at_node setAttribute table $foreigntable
	# wieder eine Voraussetzung (nicht modular)
	serviceSEdit addElem $at_node $t_node
	# generiere Assoziation Collection
	set t_node  [$domNode selectNodes \
			 descendant::table\[@name=\"$foreigntable\"\]]
	
	set ac_node [[$domNode ownerDocument] createElement assoziationcollection]
	$ac_node setAttribute name ${primtable}$foreigntable
	$ac_node setAttribute type reference
	set p_node [[$domNode ownerDocument] createElement participant]
	$p_node setAttribute reference $primattr
	$p_node setAttribute table $primtable
	$ac_node appendChild $p_node
	serviceSEdit addElem $ac_node $t_node
    }
}
Class RE_heritage -superclass ReverseEngeneering
RE_heritage instproc search {} {
    [self] instvar schema_node textwin
    # Finde alle Tabellen mit gleichnammigen Primrschlssel-Attribut 
    # Finde Alle Ein Attributige Primrschlssel
    foreach table [$schema_node selectNodes child::table] {
	set primattr [tdom_getprimattr $table]
	set tname [$table getAttribute name]
	#puts "PA $primattr $tname"
	if {[llength $primattr]==1} {
	    lappend obj_prims($primattr) $tname
	}
    }
    foreach aname [array names obj_prims] {
	if {[llength $obj_prims($aname)]>1} {
	    RE_heritage::Executor [self]::[Object autoname erExecutor] \
		-primattr $aname \
		-textwin $textwin \
		-htables $obj_prims($aname) \
	    }
    }
    next
}
Class RE_heritage::Executor -parameter {primattr roottable htables textwin}
RE_heritage::Executor instproc init {args} {
    [self] instvar primattr htables check_apply roottable check_tables textwin
    set check_apply 1
    global lan
    set wpost [namespace tail [self]]
    checkbutton $textwin.c$wpost -text $lan(re_info.acceptheritage) -variable [self]::check_apply
    $textwin window create end -window $textwin.c$wpost
    $textwin insert end " $lan(re_info.part_tables):\n"
    foreach table $htables {
	set check_tables($table) 1
	$textwin insert end "\t"
	checkbutton $textwin.ct_$table -text $lan(re_info.accepttable) -variable [self]::check_tables($table)
	$textwin window create end -window $textwin.ct_$table
	$textwin insert end "$lan(re_info.table) \"$table\"\n"
    }
    eval tk_optionMenu $textwin.op$wpost [self]::roottable $htables
    $textwin insert end "\t$lan(re_info.parent_table) :"
    $textwin window create end -window $textwin.op$wpost
    $textwin insert end "\n"
}
RE_heritage::Executor instproc apply {} {
    [self] instvar primattr check_apply roottable check_tables
    if {$check_apply} {
	# !!! Voraussetzung parent hat Attribut schema_node
	set domNode [[[self] info parent] set schema_node]
	set htables {}
	foreach ctable [array names check_tables] {
	    if {$check_tables($ctable)} {
		lappend htables $ctable
	    }
	}
	set t_node  [$domNode selectNodes \
			 descendant::table\[@name=\"$roottable\"\]]
	if {[lcontain $htables $roottable] && [llength $htables]>1} {
	    foreach ctable $htables {
		if {$ctable==$roottable} { continue }
		set p_node  [$domNode selectNodes \
				 descendant::table\[@name=\"$ctable\"\]]
		$p_node setAttribute type child
		$p_node setAttribute inheritkey $primattr
		serviceSEdit moveElem $p_node $t_node
		ProgGui applyChanges $p_node
	    }
	}
    }
}
Class RE_assoziation_rel -superclass ReverseEngeneering
RE_assoziation_rel instproc search {} {
    [self] instvar schema_node textwin
    # Untersuche die Referenzen (Frendschlssel) in den Tabellen
    # Hat die Tabelle keinen eigenen Primrschlssel oder alle 
    # Fremdschlssel-Attr  zu 
    # Primrschlssel gehren dann modelliert
    # die Tabelle eine Relation
    # Fasse zwei 1:n und 1:m zu n:m Beziehung

    foreach table [$schema_node selectNodes child::table\[@type!=\"relationship\"\]] {
	set ass_col [$table selectNodes child::assoziationcollection]
	if {[llength $ass_col]<=1} { continue }
	set primattr [tdom_getprimattr $table]
	set tname [$table getAttribute name]
	set references  [$table selectNodes child::assoziationcollection/child::participant/attribute::reference]
	set attributen [$table selectNodes {child::node()[not(self::table)]/descendant-or-self::attr/attribute::name}]
	# Kein Primrschlssel
	# Oder gehren alle Fremdschlsel-Attr zu Prim-Attributen
	if {$primattr=="" || [minusMenge $references $primattr]==""} {
	    RE_assoziation_rel::Executor [self]::[Object autoname erExecutor] \
		-references $references \
		-reftables [$table selectNodes child::assoziationcollection/child::participant/attribute::table] \
		-ass_attr [minusMenge $attributen $references] \
		-textwin $textwin \
		-table $tname
	}
    }
    next
}
Class RE_assoziation_rel::Executor -parameter {references reftables ass_attr table textwin}
RE_assoziation_rel::Executor instproc init {args} {
    [self] instvar check_apply references reftables ass_attr table textwin
    set check_apply 1
    global lan
    set wpost [namespace tail [self]]
    checkbutton $textwin.c$wpost -text $lan(re_info.accept) -variable [self]::check_apply
    $textwin window create end -window $textwin.c$wpost
    $textwin insert end " $lan(re_info.relation_table) \"$table\" $lan(re_info.degree): [llength $references]\n\t $lan(re_info.foreignkey) \"[join $references ,]\"\n\t$lan(re_info.asstables) \"[join $reftables ,]\"\n\t $lan(re_info.assattributes) \"[join $ass_attr ,]\"\n"
}
RE_assoziation_rel::Executor instproc apply {} {
    [self] instvar check_apply references reftables ass_attr table
    if {$check_apply} {
	set domNode [[[self] info parent] set schema_node]
	set t_node  [$domNode selectNodes \
			 descendant::table\[@name=\"$table\"\]]
	$t_node setAttribute type relationship
	ProgGui applyChanges $t_node
	set ass_col [$t_node selectNodes child::assoziationcollection]
	set fac_node [lindex $ass_col 0]
	#ndern von Collenction name
	$fac_node setAttribute name [$t_node getAttribute name]
	serviceSEdit newName $fac_node
	$fac_node setAttribute type reftable
	foreach ac_node [lrange $ass_col 1 end] {
	    foreach p_node [$ac_node selectNodes child::participant] {
		$ac_node removeChild $p_node
		$fac_node appendChild $p_node
	    }
	    serviceSEdit deleteElem $ac_node
	} 
	foreach otable $reftables  {
	    set tg_node [$domNode selectNodes \
			 descendant::table\[@name=\"$otable\"\]/child::assoziationtarget\[@table=\"$table\"\]]
	    $tg_node setAttribute name $table
	    serviceSEdit newName $tg_node
	} 
    }
}
Class RE_aggregation -superclass ReverseEngeneering
RE_aggregation instproc search {} {
    [self] instvar schema_node textwin
    # Untersuche die Referenzen (Frendschlssel) in den Tabellen


    foreach table [$schema_node selectNodes child::table\[@type!=\"relationship\"\]] {
	set ass_col [$table selectNodes child::assoziationcollection]
	if {[llength $ass_col]!=1} { continue }
	# Primrschlssel werden nicht bercksichtigt
	# set primattr [tdom_getprimattr $table]
	#set tname [$table getAttribute name]
	set container  [$table selectNodes child::assoziationcollection/child::participant/attribute::table]
	if {[llength $container]!=1} { continue }
	RE_aggregation::Executor [self]::[Object autoname erExecutor] \
	    -container $container \
	    -table $table \
	    -textwin $textwin \
    }
    next
}
Class RE_aggregation::Executor -parameter {container table textwin}
RE_aggregation::Executor instproc init {args} {
    [self] instvar check_apply container table textwin
    set check_apply 0
    global lan
    set wpost [namespace tail [self]]
    checkbutton $textwin.c$wpost -text $lan(re_info.accept) -variable [self]::check_apply
    $textwin window create end -window $textwin.c$wpost
    $textwin insert end " $lan(re_info.element) [$table getAttribute name] $lan(re_info.container): $container\n"
}
RE_aggregation::Executor instproc apply {} {
    [self] instvar check_apply container table
    if {$check_apply} {
	set domNode [[[self] info parent] set schema_node]
	set tname [$table getAttribute name]
	set reference [$table selectNodes child::assoziationcollection/child::participant\[@table=\"$container\"\]/attribute::reference]
	set col_table [$domNode selectNodes descendant::table\[@name=\"$container\"\]]
	# Entferne die Assoziatione Collection und Target
	serviceSEdit deleteElem [$table selectNodes \
				     child::assoziationcollection\[child::participant/attribute::table=\"$container\"\]]
	serviceSEdit deleteElem [$col_table selectNodes \
				     child::assoziationtarget\[@table=\"$tname\"\]]
	# Neue Attributte bei Element-Tabelle setzten und einbetten
	$table setAttribute reference $reference
	serviceSEdit moveElem $table $col_table
    }
}
